#include "preprocessorview.h"

#include <QStringList>
#include <KTextEditor/View>
#include <KTextEditor/Document>
#include <KTextEditor/MarkInterface>
#include <KTextEditor/Editor>
#include <KLocale>
#include <interfaces/icore.h>
#include <interfaces/idocumentcontroller.h>
#include <interfaces/idocument.h>
#include <interfaces/iprojectcontroller.h>
#include <interfaces/iproject.h>
#include <project/projectmodel.h>
#include <project/interfaces/ibuildsystemmanager.h>
#include <util/commandexecutor.h>

using namespace KDevelop;
using namespace KTextEditor;

PreprocessorView::PreprocessorView(QObject* parent, bool showIncludedFiles) : QObject(parent), state(showIncludedFiles ? WithIncludedFiles : InFile), document (KTextEditor::editor ("katepart")->createDocument(this)), view(document->createView (0))
{
		view->setWindowTitle(i18n("Preprocessor Output"));
		view->setWindowFlags(Qt::Window);
		view->setAttribute(Qt::WA_DeleteOnClose);
		connect(view, SIGNAL(destroyed()), this, SLOT(deleteMe()));
		
		document->setHighlightingMode("cpp");
		IDocument *doc;
		if((doc = ICore::self()->documentController()->activeDocument()) == 0)
				return;
		QString command = "g++ -E " + (file = doc->url().path());
		QString builddir;
		if(IProject *project = ICore::self()->projectController()->findProjectForUrl(doc->url()))
		{
				if(IBuildSystemManager *buildsystem = project->buildSystemManager())
				{
						builddir = buildsystem->buildDirectory(project->projectItem()).path();
						foreach(KUrl include, buildsystem->includeDirectories(project->projectItem()))
								command += " -I" + include.url();
				}
		}
		if(builddir == "")
		{
				builddir = doc->url().path();
				size_t x = builddir.lastIndexOf('/');
				if(x < 1)
						builddir = ".";
				else
						builddir.remove(x, builddir.length() - x);
		}
		command += " -I.";
		CommandExecutor *cmd = new CommandExecutor (command, this);
		connect(cmd, SIGNAL(receivedStandardError(const QStringList&)),
						this, SLOT(addLines(const QStringList&) ) );
		connect(cmd, SIGNAL(receivedStandardOutput(const QStringList&)),
						this, SLOT(addErrorLines(const QStringList&) ) );
		cmd->setWorkingDirectory( builddir );
		connect( cmd, SIGNAL( failed() ), this, SLOT( failed() ) );
		connect( cmd, SIGNAL( completed() ), this, SLOT( completed() ) );
		cmd->start();
}

PreprocessorView::~PreprocessorView()
{
		if(view)
				delete view;
}

void PreprocessorView::deleteMe()
{
		view = 0;
		delete this;
}

void PreprocessorView::failed()
{
		document->insertLine(document->lines(), i18n("*FAILED*"));
		qobject_cast<MarkInterface*>(document)->setMark(document->lines() - 1, MarkInterface::Execution | MarkInterface::Error);
}

void PreprocessorView::completed()
{
		document->insertLine(document->lines(), i18n("*SUCCEEDED*"));
		qobject_cast<MarkInterface*>(document)->setMark(document->lines() - 1, MarkInterface::Execution);
}

void PreprocessorView::addErrorLines(const QStringList& list)
{
		MarkInterface *mi = qobject_cast<MarkInterface*>(document);
		foreach(const QString& line, list)
		{
				document->insertLine(document->lines(), line);
				mi->setMark(document->lines() - 1, MarkInterface::Warning);
		}
}

inline bool matching(QString& line, const QString& file)
{
		int pos = line.indexOf(' ');
		bool match = false;
		if(pos > 0)
		{
				line.remove(0, pos);
				pos = line.indexOf(' ');
				if(pos > 0)
				{
						line.remove(0, pos);
						if(line.length() > 2)
						{
								line.remove(0, 1);
								if(line.startsWith(file) && line.length() > file.length() && line[file.length()] == '"')
										match = true;
						}
				}
		}
		return match;
}

void PreprocessorView::addLines(const QStringList& list)
{
		if(state == WithIncludedFiles)
		{
				foreach(QString line, list)
						document->insertLine(document->lines(), line);
		}
		else
		{
				foreach(QString line, list)
				{
						if(state == InFile)
						{
								if(line[0] == '#')
								{
										if(!matching(line, file))
												state = InAnotherFile;
								}
								else
										document->insertLine(document->lines(), line);
						}
						else
						{
								if(line[0] == '#' && matching(line, file))
										state = InFile;
						}
				}
		}
}

#include "preprocessorview.moc"
